import enum


class ParamDirection(enum.Flag):
    """Use read for readOnly, write for writeOnly; read+write if unset"""

    read = enum.auto()
    write = enum.auto()


class ParamLocation(enum.Enum):
    cookie = 'cookie'
    header = 'header'
    path = 'path'
    query = 'query'


ParamPlacement = ParamLocation



from typing import Optional

from .mime import find_mime
from .http_consts import MIME_JSON
from .response import ResponseMap


def get_accept_header(response_map: Optional[ResponseMap], global_response_map: Optional[ResponseMap]) -> Optional[str]:
    all_mime_types = {
        mime
        for rmap in [response_map, global_response_map]
        if rmap is not None
        for mime_map in rmap.values()
        for mime in mime_map.keys()
    }
    return find_mime(all_mime_types, MIME_JSON)



import logging
from typing import Optional, Any, TypeVar

import httpx
import pydantic.json

from ._params import process_params
from .pydantic_utils import to_model
from .request import get_accept_header
from .response import _handle_response, ResponseMap
from .http_consts import CONTENT_TYPE, MIME_JSON, ACCEPT

T = TypeVar('T')

logger = logging.getLogger(__name__)


class ApiBase:
    def __init__(
            self, client: httpx.AsyncClient, global_response_map: Optional[ResponseMap] = None
    ):
        self._client = client
        self._global_response_map = global_response_map

    async def __aenter__(self) -> 'ApiBase':
        await self._client.__aenter__()
        return self

    async def __aexit__(self, __exc_type=None, __exc_value=None, __traceback=None, ) -> Optional[bool]:
        return await self._client.__aexit__(__exc_type, __exc_value, __traceback)

    async def aclose(self) -> None:
        await self.__aexit__()

    async def _request(
            self,
            method: str,
            url: str,
            param_model: Optional[pydantic.BaseModel] = None,
            request_body: Any = None,
            response_map: Optional[ResponseMap] = None,
            auth: Optional[httpx.Auth] = httpx.USE_CLIENT_DEFAULT,
    ) -> T:
        request = self._build_request(method, url, param_model, request_body, response_map)

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("%s", f'{request.method} {request.url} {request.headers}')

        response = await self._client.send(request, auth=auth)

        return self._handle_response(response, response_map)

    def _build_request(
            self,
            method: str,
            url: str,
            param_model: Optional[pydantic.BaseModel],
            request_body: Any,
            response_map: Optional[ResponseMap],
    ) -> httpx.Request:
        if param_model:
            params, headers, cookies = process_params(param_model)
        else:
            params = cookies = None
            headers = httpx.Headers()

        if request_body is not None:
            headers[CONTENT_TYPE] = MIME_JSON

        if (accept := get_accept_header(response_map, self._global_response_map)) is not None:
            headers[ACCEPT] = accept

        if not isinstance(request_body, pydantic.BaseModel) and request_body is not None:
            request_body = to_model(request_body)

        content = (
            request_body.json(by_alias=True, exclude_unset=True, exclude_defaults=True)
            if request_body is not None
            else None
        )

        return self._client.build_request(
            method, url, content=content, params=params, headers=headers, cookies=cookies,
        )

    def _handle_response(
            self, response: httpx.Response, response_map: Optional[ResponseMap]
    ):
        return _handle_response(response, response_map, self._global_response_map)



import enum
import uuid
from collections.abc import Generator
from enum import Enum, unique

import httpx
import pydantic

from .absent import ABSENT
from .params import ParamLocation


@unique
class ParamStyle(Enum):
    matrix = 'matrix'
    label = 'label'
    form = 'form'
    simple = 'simple'
    spaceDelimited = 'spaceDelimited'
    pipeDelimited = 'pipeDelimited'
    deepObject = 'deepObject'


default_style = {
    ParamLocation.cookie: ParamStyle.form,
    ParamLocation.header: ParamStyle.simple,
    ParamLocation.path: ParamStyle.simple,
    ParamLocation.query: ParamStyle.form,
}


def get_style(param: pydantic.fields.ModelField) -> ParamStyle:
    # allowed styles:
    # path: matrix, label, simple
    # query: form, spaceDelimited, pipeDelimited, deepObject
    # cookie: form,
    # header: simple

    location = param.field_info.extra['in_']
    return param.field_info.extra.get('style', default_style[location])


def process_params(model: pydantic.BaseModel) -> (httpx.QueryParams, httpx.Headers, httpx.Cookies):
    containers = {
        ParamLocation.cookie: [],
        ParamLocation.header: [],
        ParamLocation.query: [],
    }

    for attr_name, param in model.__fields__.items():
        param: pydantic.fields.ModelField
        value = getattr(model, attr_name)
        if value is ABSENT:
            continue

        param_name = param.alias
        placement = param.field_info.extra['in_']
        if placement is ParamLocation.path:
            continue

        style = get_style(param)

        value = [(param_name, value) for value in serialize_param(value, style, param.field_info.extra.get('explode'))]

        containers[placement].extend(value)

    return (
        httpx.QueryParams(containers[ParamLocation.query]),
        httpx.Headers(containers[ParamLocation.header]),
        httpx.Cookies(containers[ParamLocation.cookie])
    )


def serialize_param(value, style: ParamStyle, explode_list: bool) -> Generator[str, None, None]:
    if value is None:
        return
    elif isinstance(value, str):
        yield value
    elif isinstance(value, (
            int,
            float,
            uuid.UUID,
    )):
        yield str(value)
    elif isinstance(value, enum.Enum):
        yield value.value
    elif isinstance(value, list):
        values = [
            serialized
            for val in value
            for serialized in serialize_param(val, style, explode_list)]
        if explode_list:
            # httpx explodes lists, so just pass it thru
            yield from values
        else:
            yield ','.join(values)

    else:
        raise NotImplementedError(value)



from collections.abc import Sequence
from typing import Optional

from mimeparse import best_match


def find_mime(supported_mimes: Optional[Sequence[str]], search_mime: str) -> Optional[str]:
    if supported_mimes is None or len(supported_mimes) == 0:
        return None
    match = best_match(supported_mimes, search_mime)
    return match if match != '' else None



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['lapidary_base']

package_data = \
{'': ['*']}

install_requires = \
['httpx[http2]>=0.23.0,<0.24.0',
 'pydantic>=1.10.2,<2.0.0',
 'python-mimeparse>=1.6.0,<2.0.0']

setup_kwargs = {
    'name': 'lapidary-base',
    'version': '0.5.0',
    'description': 'Base classes for clients generated by Lapidary project',
    'long_description': 'Base classes for [Lapidary](https://github.com/pytohon-lapidary/lapidary) async OpenAPI client generator.\n',
    'author': 'Raphael Krupinski',
    'author_email': 'rafalkrupinski@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/python-lapidary/lapidary-base',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from typing import Optional, Generator

import httpx
import pydantic

from .params import ParamLocation

PageFlowGenT = Generator[httpx.Request, httpx.Response, None]


class ApiKeyAuth(httpx.Auth, pydantic.BaseSettings):
    api_key: str
    name: str
    placement: ParamLocation

    def auth_flow(self, request: httpx.Request) -> PageFlowGenT:
        value = self.api_key
        if self.placement is ParamLocation.header:
            request.headers[self.name] = value
        elif self.placement is ParamLocation.query:
            request.url.params[self.name] = value
        elif self.placement is ParamLocation.cookie:
            # TODO
            raise NotImplementedError(ParamLocation.cookie)
        else:
            raise ValueError(self.placement)
        yield request


class HTTPAuth(ApiKeyAuth):
    def __init__(self, scheme: str, token: str, bearer_format: Optional[str] = None):
        value_format = 'Bearer {token}' if scheme.lower() == 'bearer' else bearer_format
        super().__init__(
            api_key=value_format.format(token=token),
            name='Authorization',
            placement=ParamLocation.header,
        )
